from album.api import Album


def album_package(album_instance: Album, args):
    from album.package.build_executable import run
    run(album_instance, args)


def create_executable_parser(parser):
    p = parser.create_command_parser('package', album_package, 'Build an package which installs the solution.')
    p.add_argument('--solution', type=str, help='path of the solution file')
    p.add_argument('--output_path', type=str, required=True, help='Path where the package solution should be written to.')



import errno
import json
import logging
import os.path
import platform
import shutil
import stat
import subprocess
import sys
from pathlib import Path

import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from album.package.install_mamba import install_mamba


# default python version for creation of an environment, in which the album-gui gets installed. 
DEFAULT_PYTHON_VERSION = "3.10"


def is_downloadable(url):
    """Shows if url is a downloadable resource."""
    with _get_session() as s:
        h = s.head(url, allow_redirects=True)
        header = h.headers
        content_type = header.get("content-type")
        if "html" in content_type.lower():
            return False
        return True


def _get_session():
    s = requests.Session()
    retry = Retry(connect=3, backoff_factor=0.5)

    adapter = HTTPAdapter(max_retries=retry)

    s.mount("http://", adapter)
    s.mount("https://", adapter)

    return s


def _request_get(url):
    """Get a response from a request to a resource url."""
    with _get_session() as s:
        r = s.get(url, allow_redirects=True, stream=True)

        if r.status_code != 200:
            raise ConnectionError("Could not connect to resource %s!" % url)

        return r


def download_resource(url, path):
    """Downloads a resource given its url."""
    path = Path(path)

    if not is_downloadable(url):
        raise AssertionError('Resource "%s" not downloadable!' % url)

    r = _request_get(url)
    with open(path, "wb") as f:
        for chunk in r:
            f.write(chunk)

    return path


def _install_missing_gui(album_base_path, mamba_base_path, album_env_path, album_gui_url, solution):
    """Function to install album gui if only the album env is present"""
    # install album gui and if needed mamba
    print("Found album, but not album gui. Installing album gui...")
    # check for mamba
    mamba_exe = get_mamba_exe(mamba_base_path)
    if not Path(mamba_exe).is_file():
        mamba_exe = install_mamba(album_base_path, mamba_base_path)

    if check_for_installed_album(album_env_path, mamba_exe):
        cmd = subprocess.run([mamba_exe, "install", "-p", album_env_path, "-c", "conda-forge", album_gui_url],
                             capture_output=True)
        if cmd.returncode == 0:
            print("Successfully installed album gui.")
        else:
            print("There was an error installing album gui: %s" % cmd.stderr)
    else:
        _install_album_full(album_base_path, mamba_base_path, album_env_path, album_gui_url, solution)


def _install_album_full(album_base_path, mamba_base_path, album_env_path, album_gui_url, solution):
    """
        Function to install album, album gui, a solution if passed and micromamba if it is not already installed
    """

    if (solution == "None") and Path(album_env_path).is_dir():
        print("Removing old album environment...")
        force_remove(album_env_path)

    # install album and album gui and if needed micromamba
    if not Path(album_base_path).is_dir():
        Path(album_base_path).mkdir()

    mamba_exe = get_mamba_exe(mamba_base_path)
    if not Path(mamba_exe).is_file():
        print("Mamba command not available. Installing Micromamba into %s..." % str(mamba_base_path))
        if not Path(mamba_base_path).is_dir():
            Path(mamba_base_path).mkdir()

        install_mamba(album_base_path, mamba_base_path)

    print("Installing album into %s..." % album_env_path)

    album_env_install_call = [mamba_exe, 'create', '--yes', '-p', album_env_path, '-c', 'conda-forge', f"python={DEFAULT_PYTHON_VERSION}", album_gui_url]
    print("installing album via %s.." % " ".join(album_env_install_call))
    a = subprocess.run(album_env_install_call, check=True)

    if a.returncode == 0:
        print("Successfully installed album.")
        if solution != "None":
            print("Installing the solution...")
            install_solution(mamba_exe, album_env_path, solution)
    else:
        print("An error occurred installing album: %s" % a.stderr)
        sys.exit()
    return True


def check_for_installed_album(album_env_path, mamba_exe):
    """Function to check if there is already an album environment in a specific directory"""
    try:
        cmd = subprocess.run([mamba_exe, "run", "-p", album_env_path, "album", "index"], capture_output=True)
        if cmd.stderr == '':
            return True
        else:
            return False
    except Exception as e:
        if Path(album_env_path).is_dir():
            answer = input("There seems to be a broken album environment at %s. "
                           "Do you want to delete it to be able to install a new album environment? (y/n)" % album_env_path)
            while True:
                if answer == 'y' or answer == 'yes':
                    print("Deleting broken album environment...")
                    shutil.rmtree(album_env_path)
                    break
                elif answer == 'n' or answer == 'no':
                    break
                else:
                    answer = input("Invalid choice please enter y/yes or n/no.")
        return False


def check_for_installed_gui(album_env_path, mamba_exe):
    """check if album gui is installed in the album environment"""
    try:
        cmd = subprocess.run([mamba_exe, "run", "-p", album_env_path, "album", "gui", "-h"], capture_output=True)
        if cmd.stderr == '':
            return True
        else:
            return False
    except Exception:
        return False


def install_solution(mamba_exe, album_env_path, solution):
    """Function to install an album solution"""
    try:
        _ = subprocess.run([mamba_exe, "run", "-p", album_env_path, "album", "install", str(solution)],
                           capture_output=True)
    except Exception as e:
        raise RuntimeError("An error occurred installing the solution") from e


def check_for_solution(mamba_exe, album_env_path, coordinates):
    """Function to check if the passed solution is installed in the album collection"""
    try:
        cmd = subprocess.run([mamba_exe, "run", "-p", album_env_path, "album", "index", '--json'],
                             capture_output=True)
        check_solutions_dict(get_solution_list_json(cmd.stdout), coordinates)
    except Exception:
        return False


def get_solution_list_json(json_str):
    """Function extracts a list of solution dictionaries from the output of an album index --json call"""
    json_str = json.loads(json_str.decode())
    # list
    catalog_list = json_str['catalogs']
    # list of catalogs_list of solution_dicts
    catalog_list_of_solution_lists = []
    # list of dict containing dicts as values
    solution_dicts_list = []

    if isinstance(catalog_list, list):
        for cat_dict in catalog_list:
            catalog_list_of_solution_lists.append(cat_dict['solutions'])
    for solution_list in catalog_list_of_solution_lists:
        for solution in solution_list:
            solution_dicts_list.append(solution)

    return solution_dicts_list


def check_solutions_dict(solution_dicts, coordinates):
    """Checks for a specific solution if it is installed, via the output of album index --json"""
    group = coordinates.split(':')[0]
    name = coordinates.split(':')[1]
    version = coordinates.split(':')[2]

    # Go through the list of dictionaries and checking there setup method for the coordinates and then check the internal
    # string for the installation status
    for solution in solution_dicts:
        internal = solution['internal']
        setup = solution['setup']
        installed = internal['installed']
        tmp_group = setup['group']
        tmp_name = setup['name']
        tmp_version = setup['version']
        if (group == tmp_group) and (name == tmp_name) and (version == tmp_version):
            if installed == 1:
                print("solution already installed")
                return True
            else:
                return False

    return False


def force_remove(path, warning=True):
    """Function to force remove a specific file/directory"""
    path = Path(path)
    if path.exists():
        try:
            if path.is_file():
                try:
                    path.unlink()
                except PermissionError:
                    handle_remove_readonly(os.unlink, path, sys.exc_info())
            else:
                shutil.rmtree(
                    str(path), ignore_errors=False, onerror=handle_remove_readonly
                )
        except PermissionError as e:
            logging.warning("Cannot delete %s." % str(path))
            if not warning:
                raise e


def handle_remove_readonly(func, path, exc):
    """Changes readonly flag of a given path."""
    excvalue = exc[1]
    if func in (os.rmdir, os.remove, os.unlink) and excvalue.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
        func(path)
    else:
        raise


def create_shortcut(mamba_exe, album_env_path):
    """Creates a shortcut for the album gui on the users desktop"""
    subprocess.run([mamba_exe, 'run', '-p', album_env_path, 'album', 'add-shortcut'], check=True)


def get_album_base_path():
    """returns the base directory of the album collection"""
    return Path.home().joinpath('.album')


def get_album_env_path(album_base_path):
    """returns the path to the album environment"""
    return Path(album_base_path).joinpath("envs", "album")


def get_mamba_exe(mamba_base_path):
    """returns the path to the micromamba executable"""
    if platform.system() == 'Windows':
        return str(Path(mamba_base_path).joinpath('Library', 'bin', 'micromamba.exe'))
    else:
        return str(Path(mamba_base_path).joinpath('bin', 'micromamba'))


def read_solution_info():
    """Function to read the information of the solution which should be installed.
    Returns path to the solution and it's coordinates. solution_info.txt is used to pass the information about
    the solution into the executable without destroying the import structure ob pyinstaller"""
    with open(Path(os.path.realpath(__file__)).parent.joinpath('solution_info.txt'), 'r') as file:
        sol_info = file.read()
    return sol_info.split('\n')


def main():
    album_base_path = str(get_album_base_path())
    mamba_base_path = str(Path(album_base_path).joinpath("micromamba"))
    mamba_exe = get_mamba_exe(mamba_base_path)
    album_env_path = str(get_album_env_path(album_base_path))
    album_gui_url = "album-gui=0.4.*"
    solution_info = read_solution_info()
    solution_path = solution_info[0]
    coordinates = solution_info[1]
    full_install = False

    if not Path(album_base_path).is_dir():
        Path(album_base_path).mkdir()

    print("Checking for micromamba installation...")
    if not Path(mamba_exe).is_file():
        print("Mamba command not available. Installing micromamba into %s..." % mamba_base_path)
        install_mamba(album_base_path, mamba_base_path)

    if not check_for_installed_album(album_env_path, mamba_exe) or (solution_path == "None"):
        full_install = _install_album_full(album_base_path, mamba_base_path, album_env_path, album_gui_url, solution_path)

    if not full_install:
        print("Checking for album gui installation...")
        if not check_for_installed_gui(album_env_path, mamba_exe):
            _install_missing_gui(album_base_path, mamba_base_path, album_env_path, album_gui_url, solution_path)

        if solution_path != "None":
            print("Checking if the solution is installed...")
            if not check_for_solution(mamba_exe, album_env_path, solution_path):
                print("Installing the solution...")
                install_solution(mamba_exe, album_env_path, solution_path)

    print("Creating shortcut..")
    create_shortcut(mamba_exe, album_env_path)

    print("Installation successfully finished.")


if __name__ == '__main__':
    main()



import logging
import os
import platform
import re
from pathlib import Path

import PyInstaller.__main__
import pkg_resources

from album.api import Album
from album.runner.album_logging import get_active_logger
from album.package.install_all import force_remove

# remove all handlers from root logger. Necessary because PyInstaller changes root logger configuration.
for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)
    handler.close()


def fill_placeholder(output_path, solution_path, coordinates):
    """Fills the path and the coordinates of the passed solution into a cache file which gets copied into the executable
    so the information is availabe in the executable without destroying pyinstallers import structure"""
    with open(pkg_resources.resource_filename('album.package', 'solution_info.txt'), 'r') as file:
        template_str = file.read()

    # make windowspaths readable by escaping backslashes
    tmp = re.sub(r"\\", r"\\\\", solution_path)

    template_str = re.sub("<solution_placeholder>", tmp, template_str)
    template_str = re.sub(r"\\", r"\\\\", template_str)
    with open(Path(output_path).joinpath('solution_info.txt'), 'w') as file:
        file.write(template_str)
        file.write("\n")
        file.write(coordinates)


def create_pyinstaller_params(args, name_param):
    """creates the parameter for the pyinstaller call"""
    work_dir = Path(args.output_path).joinpath('build')
    workpath_param = '--workpath=%s' % str(work_dir)
    exe_path_param = '--distpath=%s' % str(args.output_path)
    spec_path = Path(args.output_path)
    spec_path_param = '--specpath=%s' % str(args.output_path)
    log_level_param = '--log-level=ERROR'
    script_path = str(pkg_resources.resource_filename('album.package', 'install_all.py'))
    install_mamba_script = pkg_resources.resource_filename('album.package', 'install_mamba.py')
    install_mamba_param = '--add-data=%s%s%s' % (install_mamba_script, os.pathsep, '.')
    solution_info_path = str(Path(args.output_path).joinpath('solution_info.txt'))
    solution_info_param = '--add-data=%s%s%s' % (solution_info_path, os.pathsep, '.')

    pyinstaller_params = [script_path, '--onefile', name_param, exe_path_param, install_mamba_param,
                          solution_info_param, workpath_param, spec_path_param, log_level_param]

    hidden_imports = ["pkg_resources"]
    for hidden_import in hidden_imports:
        pyinstaller_params.append("--hidden-import=%s" % hidden_import)

    return pyinstaller_params


def clean_up(output_path, name):
    force_remove(Path(output_path).joinpath('build'))
    force_remove(Path(output_path).joinpath('solution_info.txt'))
    force_remove(Path(output_path).joinpath('%s.spec' % name))


def run(album_instance: Album, args):
    """builds an executable which installs album and inf wanted an album solution"""
    try:
        name = ""
        if not Path(args.output_path).is_dir():
            Path(args.output_path).mkdir()

        if args.solution is None:
            get_active_logger().info("Build an executable which installs album.")
        if (args.solution is not None) and (not (Path(args.solution).exists())):
            args.solution = album_instance.resolve(str(args.solution)).path()
        if args.solution is not None:
            get_active_logger().info("Build an executable which installs the solution.")
            coordinates = album_instance.resolve(str(args.solution)).coordinates()
            get_active_logger().info(
                "solution: %s at %s" % (coordinates, args.solution))
            if platform.system() == "Darwin":
                name = "%s_%s_%s_installer.app" % (coordinates.group(), coordinates.name(),
                                                   str(coordinates.version()).replace('.', '_'))
                name_param = "--name=%s_%s_%s_installer.app" % (
                    coordinates.group(), coordinates.name(), str(coordinates.version()).replace('.', '_'))
            else:
                name = "%s_%s_%s_installer" % (coordinates.group(), coordinates.name(),
                                               str(coordinates.version()).replace('.', '_'))
                name_param = "--name=%s_%s_%s_installer" % (
                    coordinates.group(), coordinates.name(), str(coordinates.version()).replace('.', '_'))
            fill_placeholder(str(args.output_path), str(args.solution), coordinates.__str__())
        else:
            if platform.system() == "Darwin":
                name = "album_installer.app"
                name_param = "--name=album_installer.app"
            else:
                name = "album_installer"
                name_param = "--name=album_installer"
            fill_placeholder(str(args.output_path), str(args.solution), "None")

        get_active_logger().info("--output_path: %s" % args.output_path)

        pyinstaller_params = create_pyinstaller_params(args, name_param)

    except Exception as e:
        clean_up(args.output_path, name)
        raise RuntimeError("Unexpected Error when preparing the executable build %s." % e) from e

    try:
        PyInstaller.__main__.run(pyinstaller_params)
    except Exception as e:
        raise RuntimeError("PyInstaller exited with an unexpected error! %s" % e) from e
    finally:
        clean_up(args.output_path, name)





import os
import platform
import subprocess
from pathlib import Path
import zipfile
import tarfile

import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

MAMBA_LINK_LINUX_X86_64 = "https://micro.mamba.pm/api/micromamba/linux-64/1.5.6"
MAMBA_LINK_LINUX_ARM64 = "https://micro.mamba.pm/api/micromamba/linux-aarch64/1.5.6"
MAMBA_LINK_LINUX_POWER = "https://micro.mamba.pm/api/micromamba/linux-ppc64le/1.5.6"

MAMBA_LINK_MACOS_X86_64 = "https://micro.mamba.pm/api/micromamba/osx-64/1.5.6"
MAMBA_LINK_MACOS_ARM64 = "https://micro.mamba.pm/api/micromamba/osx-arm64/1.5.6"

MAMBA_LINK_WINDOWS = "https://gitlab.com/album-app/plugins/album-package/-/raw/micromamba_installer/win-64_micromamba-1.5.6-0.zip?ref_type=heads&inline=false"


def is_downloadable(url):
    """Shows if url is a downloadable resource."""
    with _get_session() as s:
        h = s.head(url, allow_redirects=True)
        header = h.headers
        content_type = header.get("content-type")
        if "html" in content_type.lower():
            return False
        return True


def _get_session():
    s = requests.Session()
    retry = Retry(connect=3, backoff_factor=0.5)

    adapter = HTTPAdapter(max_retries=retry)

    s.mount("http://", adapter)
    s.mount("https://", adapter)

    return s


def _request_get(url):
    """Get a response from a request to a resource url."""
    with _get_session() as s:
        r = s.get(url, allow_redirects=True, stream=True)

        if r.status_code != 200:
            raise ConnectionError("Could not connect to resource %s!" % url)

        return r


def download_resource(url, path):
    """Downloads a resource given its url."""
    path = Path(path)

    if not is_downloadable(url):
        raise AssertionError('Resource "%s" not downloadable!' % url)

    r = _request_get(url)
    with open(path, "wb") as f:
        for chunk in r:
            f.write(chunk)

    return path


def check_architecture():
    """checks the processor architecture of the system"""
    check = subprocess.run(['uname', '-m'], capture_output=True)
    return check.stdout.decode().rstrip()


def _download_mamba(mamba_base_path):
    """downloads micromamba"""
    if platform.system() == 'Windows':
        return _download_mamba_win(Path(mamba_base_path).joinpath('micromamba.zip'))
    elif platform.system() == 'Darwin':
        return _download_mamba_macos(Path(mamba_base_path).joinpath('micromamba.tar'))
    elif platform.system() == 'Linux':
        return _download_mamba_linux(Path(mamba_base_path).joinpath('micromamba.tar'))
    else:
        raise NotImplementedError("Your operating system is currently not supported.")


def _download_mamba_win(mamba_installer_path):
    """downloads micromamba for windows"""
    return download_resource(MAMBA_LINK_WINDOWS, mamba_installer_path)


def _download_mamba_macos(mamba_installer_path):
    """downloads micromamba for macOS depending on the processor architecture"""
    if check_architecture().__eq__('x86_64'):
        return download_resource(MAMBA_LINK_MACOS_X86_64, mamba_installer_path)
    elif check_architecture().lower().__eq__('arm64'):
        return download_resource(MAMBA_LINK_MACOS_ARM64, mamba_installer_path)
    else:
        raise NotImplementedError("There is no micromamba version for your processor architecture.")


def _download_mamba_linux(mamba_installer_path):
    """downloads micromamba for linux depending on the processor architecture"""
    if check_architecture().__eq__('x86_64'):
        return download_resource(MAMBA_LINK_LINUX_X86_64, mamba_installer_path)
    elif check_architecture().lower().__eq__('arm64'):
        return download_resource(MAMBA_LINK_LINUX_ARM64, mamba_installer_path)
    elif check_architecture().lower().__eq__('power'):
        return download_resource(MAMBA_LINK_LINUX_POWER, mamba_installer_path)
    else:
        raise NotImplementedError("There is no micromamba version for your processor architecture.")


def _unpack_mamba_win(mamba_installer, mamba_base_path):
    """unpacks the windows version of the micromamba archive"""
    with zipfile.ZipFile(Path(mamba_installer)) as zip:
        zip.extractall(Path(mamba_base_path))


def _unpack_mamba_unix(mamba_installer, mamba_base_path):
    """unpacks the micromamba archives for linux and macOS"""
    with tarfile.open(mamba_installer, 'r') as tar:
        tar.extractall(mamba_base_path)


def _set_mamba_env_vars(mamba_base_path):
    """Sets the micromamba environment variables"""
    os.environ['MAMBA_ROOT_PREFIX'] = str(mamba_base_path)
    os.environ['MAMBA_EXE'] = str(get_mamba_exe(mamba_base_path))


def _init_pwsh(mamba_base_path):
    """initializes the windows powershell for the usage of micromamba"""
    mamba_exe = get_mamba_exe(mamba_base_path)
    _ = subprocess.run([mamba_exe, 'shell', 'init', '-s', 'powershell', '-p', mamba_base_path, '-y'],
                       capture_output=True)

# Cmd.exe will not be initialised since it will break existing conda installations
#def _init_cmd_exe(mamba_base_path):
#    """initializes the windows cmd.exe for the usage of micromamba"""
#    mamba_exe = get_mamba_exe(mamba_base_path)
#    _ = subprocess.run([mamba_exe, 'shell', 'init', '-s', 'cmd.exe', '-p', mamba_base_path, '-y'],
#                       capture_output=True)


def _init_bash(mamba_base_path):
    """initializes the bash shell for the usage of micromamba"""
    mamba_exe = get_mamba_exe(mamba_base_path)
    _ = subprocess.run([mamba_exe, 'shell', 'init', '-s', 'bash', '-p', mamba_base_path],
                       capture_output=True)


def _init_zsh(mamba_base_path):
    """initializes the zsh shell for the usage of micromamba"""
    mamba_exe = get_mamba_exe(mamba_base_path)
    _ = subprocess.run([mamba_exe, 'shell', 'init', '-s', 'zsh', '-p', mamba_base_path],
                       capture_output=True)


def get_mamba_exe(mamba_base_path):
    """returns the path to the micromamba executable"""
    if platform.system() == 'Windows':
        return str(Path(mamba_base_path).joinpath('Library', 'bin', 'micromamba.exe'))
    else:
        return str(Path(mamba_base_path).joinpath('bin', 'micromamba'))


def _install_mamba_windows(mamba_base_path):
    """installs micromamba on windows"""
    mamba_installer = _download_mamba(mamba_base_path)
    _unpack_mamba_win(mamba_installer, mamba_base_path)
    _set_mamba_env_vars(mamba_base_path)
    _init_pwsh(mamba_base_path)


def _install_mamba_linux(mamba_base_path):
    """installs micromamba on linux"""
    mamba_installer = _download_mamba(mamba_base_path)
    _unpack_mamba_unix(mamba_installer, mamba_base_path)
    _set_mamba_env_vars(mamba_base_path)
    _init_bash(mamba_base_path)


def _install_mamba_macos(mamba_base_path):
    """installs micromamba on macOS"""
    mamba_installer = _download_mamba(mamba_base_path)
    _unpack_mamba_unix(mamba_installer, mamba_base_path)
    _set_mamba_env_vars(mamba_base_path)
    _init_zsh(mamba_base_path)


def install_mamba(album_base_path, mamba_base_path):
    """installs micormamba"""
    if not Path(album_base_path).exists():
        Path(album_base_path).mkdir()
    if not Path(mamba_base_path).exists():
        Path(mamba_base_path).mkdir()

    if platform.system() == 'Windows':
        _install_mamba_windows(mamba_base_path)
    elif platform.system() == 'Darwin':
        _install_mamba_macos(mamba_base_path)
    elif platform.system() == 'Linux':
        _install_mamba_linux(mamba_base_path)
    else:
        raise NotImplementedError("Your operating system is currently not supported")
